#
# rule.rb
#
#   Copyright (c) 1999,2000 Minero Aoki <aamine@dp.u-netsurf.ne.jp>
#
#   This program is free software.
#   You can distribute/modify this program under the terms of
#   the GNU Lesser General Public License version 2 or later.
#


module Racc

  class RuleTable

    def initialize( rac )
      @racc = rac
      @tokentable = rac.tokentable

      @d_token = rac.d_token
      @d_rule  = rac.d_rule
      @d_state = rac.d_state

      @rules    = []
      @finished = false
      @hashval  = 4
      @start    = nil
    end


    def register( symbol, rulearr, tempprec, act )
      rule = Rule.new(
        symbol, rulearr, act,
        @rules.size + 1,         # ID
        @hashval,                # hash value
        tempprec                 # prec
      )
      @rules.push rule

      @hashval += rulearr.size + 1
    end

    def start=( sim )
      unless @start then
        @start = sim
        true
      else
        false
      end
    end
    
    attr :start


    def init

      ### add dummy rule

      # array size must be synced with @hashval (in register.)
      temp = Rule.new(
          @tokentable.dummy,
          [ @start, @tokentable.anchor, @tokentable.anchor ],
          Action.new( '', 0 ),
          0, 0, nil )
        # id hash prec
      @rules.unshift temp
      @rules.freeze


      ### cache

      @rules.each do |rule|
        rule.symbol.heads.push rule.ptrs[0]
      end

      @tokentable.fix

      @rules.each do |rule|
        temp = nil

        rule.each_ptr do |ptr|
          tok = ptr.unref
          tok.locate.push ptr
          temp = tok if tok.term
        end

        rule.prec = temp if rule.prec.nil?
      end
    end


    def []( x )
      @rules[x]
    end

    def each_rule( &block )
      @rules.each( &block )
    end
    alias each each_rule

    def each_index( &block )
      @rules.each_index( &block )
    end

    def each_with_index( &block )
      @rules.each_with_index( &block )
    end

    def size
      @rules.size
    end

    def to_s
      "<Racc::RuleTable>"
    end

  end   # RuleTable



  class Rule

    def initialize( tok, rlarr, act, rid, hval, tprec )
      @symbol  = tok
      @rulearr = rlarr
      @action  = act.val
      @lineno  = act.lineno
      @ident   = rid
      @prec    = tprec

      @ptrs = []
      rlarr.each_index do |idx|
        @ptrs[ idx ] = LocationPointer.new( self, idx, rlarr[idx], hval + idx )
      end

      # reduce pos
      s = rlarr.size
      @ptrs[ s ] = LocationPointer.new( self, s, nil, hval + s )
    end

    attr :ident
    alias hash ident

    attr :ptrs
    attr :action
    attr :lineno
    attr :symbol
    attr :prec, true

    def ==( other )
      Rule === other and @ident == other.ident
    end

    def accept?()
      if tok = @rulearr[-1] then
        tok.anchor?
      else
        false
      end
    end

    def []( idx )
      @rulearr[idx]
    end

    def size
      @rulearr.size
    end

    def to_s
      "<Rule #{@ident}>"
    end

    def pointers
      @ptrs.dup
    end

    def toks( idx )
      @rulearr[idx]
    end

    def tokens
      @rulearr.dup
    end

    def each_token( &block )
      @rulearr.each( &block )
    end
    alias each each_token

    def each_with_index( &block )
      @rulearr.each_with_index( &block )
    end

    def each_ptr( &block )
      pmax = @ptrs.size - 1
      i = 0
      while i < pmax do
        yield @ptrs[i]
        i += 1
      end
    end

  end   # Rule



  class LocationPointer

    def initialize( rl, idx, tok, id )
      @rule   = rl
      @index  = idx
      @unref  = tok
      @ident  = id
      @reduce = tok.nil?
    end

    attr :ident
    alias hash ident

    attr :rule
    attr :index
    attr :unref

    def to_s
      sprintf( '(%d,%d %s)',
               @rule.ident, @index, reduce? ? '#' : unref.to_s )
    end
    alias inspect to_s

    def eql?( ot )
      @ident == ot.ident
    end
    alias == eql?

    def reduce?
      @unref.nil?
    end

    def head?
      @index == 0
    end

    def increment
      ret = @rule.ptrs[ @index + 1 ]
      ret or ptr_bug!
    end

    def decrement
      ret = @rule.ptrs[ @index - 1 ]
      ret or ptr_bug!
    end

    def before( len )
      ret = @rule.ptrs[ @index - len ]
      ret or ptr_bug!
    end

    private
    
    def ptr_bug!
      bug! "pointer not exist: self: #{to_s}"
    end

  end   # class LocationPointer



  class TokenTable

    include Enumerable

    def initialize( racc )
      @racc = racc

      @chk = {}
      @tokens = []
      
      @dummy   = get( :$start )
      @anchor  = get( :$end )
      @error   = get( :error )   # error token is ID 1

      @anchor.conv = 'false'
      @error.conv = 'Object.new'
    end

    attr :dummy
    attr :anchor

    def get( val )
      unless ret = @chk[ val ] then
        @chk[ val ] = ret = Token.new( val, @racc )
        @tokens.push ret
      end

      ret
    end

    def fix
      tok = nil
      @tokens.each do |tok|
        tok.set_term tok.heads.empty?
      end

      term = []
      nt = []
      i = nil
      @tokens.each do |i|
        (i.terminal? ? term : nt).push i
      end
      @tokens = term
      @nt_base = term.size
      term.concat nt

      @tokens.each_with_index do |t, i|
        t.set_ident i
      end
    end

    attr :nt_base

    def nt_max
      @tokens.size
    end

    def each( &block )
      @tokens.each &block
    end

    def each_terminal( &block )
      @tokens[ 0, @nt_base ].each( &block )
    end

    def each_nonterm( &block )
      @tokens[ @nt_base, @tokens.size - @nt_base ].each( &block )
    end


    def init
      tok = nil
      each_nonterm {|tok| tok.compute_expand }
      each_nonterm {|tok| tok.compute_svoid }
      each_nonterm {|tok| tok.compute_void }
      each_nonterm {|tok| tok.compute_first }
      each_nonterm {|tok| tok.compute_follow }
    end

  end   # TokenTable


  class Token

    Default_token_id = -1
    Anchor_token_id  = 0
    Error_token_id   = 1

    def initialize( tok, racc )
      @value   = tok
      @d_token = racc.d_token

      @ident  = nil
      @term   = nil
      @conv   = nil
      @heads  = []
      @locate = []
      @svoid  = nil
      @void   = nil
      @expand = nil
      @first  = nil
      @follow = nil

      if @value.respond_to?(:id2name) then
        @to_s   = @value.id2name
        @uneval = ':' + @value.id2name
      else
        @to_s   = @value.inspect
        @uneval = @value.inspect
      end
    end


    attr :ident
    alias hash ident

    def set_ident( id )
      bug! unless @ident.nil?
      @ident = id
    end


    attr :value
    attr :term
    attr :nterm
    attr :conv
    attr :prec,  true
    attr :assoc, true

    alias terminal? term
    alias nonterminal? nterm

    def set_term( t )
      bug! unless @term.nil?
      @term = t
      @nterm = !t
    end

    def conv=( str )
      @conv = @uneval = str
    end


    attr :heads
    attr :locate
    attr :expand
    attr :svoid
    attr :void
    attr :first
    attr :follow

    alias void_self? svoid
    alias void?      void


    def to_s
      @to_s.dup
    end
    alias inspect to_s

    def uneval
      @uneval.dup
    end


    # only on left
    def useless?
      @nterm and @locate.empty?
    end


    def compute_expand
      puts "expand> #{to_s}" if @d_token
      @expand = upd_expand( [], [] )
      @expand.compact!
      puts "expand< #{to_s}: #{@expand.join(' ')}" if @d_token
    end

    def upd_expand( ret, lock )
      if @expand then
        @expand.each {|i| ret[i.ident] = i }
        return
      end
      lock[ @ident ] = true

      tok = nil
      @heads.each do |ptr|
        ret[ ptr.ident ] = ptr
        if (tok = ptr.unref) and tok.nonterminal? then
          unless lock[ tok.ident ] then
            tok.upd_expand( ret, lock )
          end
        end
      end

      ret
    end


    def compute_svoid
      @svoid = false
      ptr = nil
      @heads.each do |ptr|
        if ptr.reduce? then
          @svoid = true
          break
        end
      end

      puts "svoid?: #{to_s}: #{@svoid}" if @d_token
    end


    def compute_void
      @void = false
      ptr = nil
      @expand.each do |ptr|
        if ptr.reduce? then
          @void = true
          break
        end
      end

      puts "void?: #{to_s}: #{@void}" if @d_token
    end


    def compute_first
      puts "first> #{to_s}" if @d_token
      @first = upd_first( [], [] )
      @first.compact!
      puts "first< #{to_s}: #{@first.join(' ')}" if @d_token
    end

    def upd_first( ret, lock )
      ptr = tok = i = nil

      if @first then
        @first.each {|i| ret[i.ident] = i }
        return
      end
      lock[@ident] = true
      if @term then
        bug! '"first" called for terminal'
      end

      @heads.each do |ptr|
        until ptr.reduce? do
          tok = ptr.unref
          if tok.terminal? then
            ret[ tok.ident ] = tok
            break
          else
            tok.upd_first( ret, lock ) unless lock[tok.ident]
            break unless tok.void?
          end

          ptr = ptr.increment
        end
      end

      ret
    end


    def compute_follow
      puts "follow> #{to_s}" if @d_token
      @follow = upd_follow( [], [] )
      @follow.compact!
      puts "follow< #{to_s}: #{@follow.join(' ')}" if @d_token
    end

    def upd_follow( ret, lock )
      if @follow then
        @follow.each {|i| ret[i.ident] = i }
        return
      end
      lock[ @ident ] = true

      ptr = tok = nil

      @locate.each do |ptr|
        while true do
          ptr = ptr.increment || ptr
          if ptr.reduce? then
            tok = ptr.rule.symbol
            tok.upd_follow( ret, lock ) unless lock[ tok.ident ]
            break
          end
          tok = ptr.unref
          if tok.terminal? then
            ret[ tok.ident ] = tok
            break
          else
            tok.first.each {|i| ret[i.ident] = i }
            break unless tok.void?
          end
        end
      end

      ret
    end

  end   # class Token

end   # module Racc
