#
# state.rb
#
#   Copyright (c) 1999,2000 Minero Aoki <aamine@dp.u-netsurf.ne.jp>
#
#   This program is free software.
#   You can distribute/modify this program under the terms of
#   the GNU Lesser General Public License version 2 or later.
#

require 'amstd/must'


module Racc

  class RaccError < StandardError; end


  class LALRstateTable

    def initialize( racc )
      @racc = racc
      @ruletable  = racc.ruletable
      @tokentable = racc.tokentable

      @d_state   = racc.d_state
      @d_reduce  = racc.d_reduce
      @d_shift   = racc.d_shift
      @d_verbose = racc.d_verbose
      @d_profile = racc.d_profile

      @states = []
      @statecache = {}

      @actions = LALRactionTable.new( @ruletable, self )
    end
    

    attr :actions

    def size
      @states.size
    end


    def inspect
      "#<StateTable size=#{@states.size}>"
    end
    alias to_s inspect

    def []( i )
      @states[i]
    end

    def each_state( &block )
      @states.each( &block )
    end
    alias each each_state

    def each_index( &block )
      @states.each_index( &block )
    end


    def init

      # add state 0
      core_to_state( [ @ruletable[0].ptrs[0] ] )

      cur = 0
      while cur < @states.size do
        build_state @states[cur]   # state is added here
        cur += 1
      end
      @states.each do |st|
        st.from_table.each do |t,arr|
          arr.compact!
        end
      end
    end

    
    def build_state( state )
      puts "build_state: #{state}" if @d_state

      bcore = {}
      rp = state.reduce_ptrs
      gt = state.goto_table

      state.closure.each do |ptr|
        if ptr.reduce? then
          rp.push ptr
        else
          tok = ptr.unref
          gt[tok] = true

          if arr = bcore[ tok ] then
            arr.push ptr.increment
          else
            bcore[ tok ] = [ ptr.increment ]
          end
        end
      end

      bcore.each do |tok, core|
        # all 'core's must be rule table order (upper to lower)
        core.sort!{|a, b| a.hash <=> b.hash }

        puts "build_state: tok=#{tok} ncore=#{core.join(' ')}" if @d_state

        dest = core_to_state( core )
        connect( state, tok, dest )
      end
    end

    def core_to_state( core )
      k = cdigest( core )
      unless dest = @statecache[k] then
        # not registered yet
        dest = LALRstate.new( @states.size, core, @racc )
        @states.push dest

        @statecache[k] = dest
        
        puts "core_to_state: create state   ID #{dest.ident}" if @d_state
      else
        if @d_state then
          puts "core_to_state: dest is cached ID #{dest.ident}"
          puts "core_to_state: dest core #{dest.core.join(' ')}"
        end
      end

      dest
    end

    def cdigest( arr )
      arr.collect {|ptr| ptr.ident }.pack( 'L*' )
    end

    def connect( from, tok, dest )
      puts "connect: #{from.ident} --(#{tok})-> #{dest.ident}" if @d_state

      # check infinite recursion
      if from.ident == dest.ident and from.closure.size < 2 then
        raise RaccError, sprintf( "Infinite recursion: state %d, with rule %d",
          from.ident, from.core[0].rule.ident )
      end

      # goto
      from.goto_table[ tok ] = dest

      # come from
      unless tmp = dest.from_table[ tok ] then
        dest.from_table[ tok ] = tmp = []
      end
      tmp[ from.ident ] = from
    end


    def resolve
      $stderr.puts "resolving #{@states.size} states" if @d_verbose

      @states.each do |state|
        state.compute_first_term
      end

      b = Time.times.utime
      slr = 0
      lalr = 0
      @states.each do |state|
        if @d_verbose and state.ident % 40 == 39 then
          $stderr.puts "end #{state.ident + 1} states"
        end

        if state.conflicting? then
          ret = state.resolve
          if ret == :slr then slr += 1 else lalr += 1 end
        end
      end

      if @d_verbose then
        e = Time.times.utime
        $stderr.puts "all resolved in #{e - b} sec: slr #{slr}, lalr #{lalr}"
      end

      # set accept

      anch = @tokentable.anchor
      init_state = @states[0].goto_table[ @ruletable.start ]
      targ_state = init_state.action[ anch ].goto_state
      acc_state  = targ_state.action[ anch ].goto_state

      acc_state.action.clear
      acc_state.goto_table.clear
      acc_state.defact = @actions.accept

      pack_states
    end


    def pack_states
      st = arr = act = nil
      i = n = s = r = nil

      @states.each do |st|
        #
        # find most frequently used reduce rule
        #
        act = st.action
        arr = Array.new( @ruletable.size, 0 )
        act.each do |t,a|
          if ReduceAction === a then
            arr[ a.ruleid ] += 1
          end
        end
        i = s = nil
        i = arr.max
        s = arr.index( i ) if i > 0

        if s then
          r = @actions.reduce(s)
          if not st.defact or st.defact == r then
            act.delete_if {|t,a| a == r }
            st.defact = r
          end
        else
          st.defact ||= @actions.error
        end
      end
    end

  end   # LALRstateTable


  class SLRerror < StandardError; end


  class LALRstate

    def initialize( sid, core, racc )
      @ident = sid
      @core = core

      @closure = nil

      @racc       = racc
      @tokentable = racc.tokentable
      @actions    = racc.statetable.actions
      @d_state    = racc.d_state
      @d_reduce   = racc.d_reduce
      @d_shift    = racc.d_shift
      @d_verbose  = racc.d_verbose
      @d_profile  = racc.d_profile

      @from_table    = {}
      @goto_table    = {}

      @first_term = {}
      @void       = nil

      @shift_toks  = []
      @reduce_ptrs = []
      @reduce_core = []

      @action = {}
      @defact = nil

      @resolve_log = []
      @rrconf      = nil
      @srconf      = nil

      init_values
    end

    def init_values
      @reduce_core = core.find_all {|ptr| ptr.reduce? }

      @closure = compute_closure( @core )

      @void = []
      t = nil
      @closure.each do |ptr|
        t = ptr.unref
        if t and t.void? then
          @void.push t
        end
      end
    end

    def compute_closure( ptrs )
      puts "closure: ptrs=#{ptrs.join(' ')}" if @d_state

      ret = []
      tok = i = nil
      ptrs.each do |ptr|
        ret[ ptr.ident ] = ptr
        tok = ptr.unref
        if tok and tok.nonterminal? then
          tok.expand.each {|i| ret[i.ident] = i }
        end
      end
      ret.compact!

      puts "closure: ret #{ret.join(' ')}" if @d_state
      ret
    end


    attr :ident
    alias hash ident

    attr :core
    attr :closure

    attr :from_table
    attr :goto_table

    attr :void
    attr :first_term

    attr :reduce_ptrs
    attr :reduce_core

    attr :action
    attr :defact, true   # default action

    attr :resolve_log

    attr :rrconf
    attr :srconf

    def inspect
      "#<state #{@ident}>"
    end
    alias to_s inspect

    def ==( oth )
      @ident == oth.ident
    end
    alias eql? ==


    def conflicting?
      if @reduce_ptrs.size > 0 then
        if @closure.size == 1 then
          #
          # reduce
          #
          @defact = @actions.reduce( @reduce_ptrs[0].rule )
        else
          #
          # conflict
          #
          return true
        end
      else
        #
        # shift
        #
        @goto_table.each do |tok, goto|
          if tok.terminal? then
            @action[ tok ] = @actions.shift( goto )
          end
        end
      end

      false
    end


    def resolve
      if @d_reduce or @d_shift then
        puts "resolving state #{@ident} -------------"
      end

      @method = :slr
      begin
        resolve_rr
        resolve_sr
      rescue SLRerror
        if @d_reduce or @d_shift then
          puts "state #{@ident}, slr fail -------------"
        end
        @action.clear
        @method = :lalr
        retry
      end

      @method
    end


    #
    #  Reduce/Reduce Conflict resolver
    #

#trap( 'INT' ) do
#puts "resolving #{$resolving_state.ident}"; $stdout.flush
#raise Exception, "stopped by SIGINT"
#end

    def resolve_rr
#$resolving_state = self
      puts "resolve_rr: state #{@ident}, #{method.id2name}" if @d_reduce

      la = act = curptr = tok = nil

      @reduce_ptrs.each do |curptr|
        puts "resolve_rr: resolving #{curptr}" if @d_reduce

        la = send( @method, curptr )
        la.each do |tok|
          act = @action[ tok ]
          if act then
            unless ReduceAction === act then
              bug! "no reduce action #{act.type} in action table"
            end
            #
            # can't resolve R/R conflict (on tok).
            #   reduce with upper rule as default
            #

            rr_conflict act.rule, curptr.rule, tok
          else
            # not conflict
            @action[ tok ] = @actions.reduce( curptr.rule )
          end
        end
      end
    end


    def slr( ptr )
      ptr.rule.symbol.follow
    end


    def lalr( ptr )
      ret = lookahead__new( ptr )
      ret.compact!
      puts "LA: state #{@ident}: #{ret.join(' ')}" if @d_reduce
      ret
    end

    def compute_first_term
      @first_term = ret = []
      t = i = nil
      @goto_table.each_key do |t|
        if t.terminal? then
          ret[t.ident] = t
        else
          t.first.each {|i| ret[i.ident] = i }
        end
      end
      ret.compact!
    end

    def lookahead__old( ptr )
      puts "la> state #{@ident},#{ptr}" if @d_reduce
      st_beg = Time.times.utime

      sim = ptr.rule.symbol
      ret = []
      gotos = {}
      new = {}
      lock = []
      goto = pt = ff = len = f = a = tmp = nt = h = nil
      d_reduce = @d_reduce

      head_state( ptr ).each_key do |f|
        tmp = f.goto_table[ sim ]
        if a = gotos[tmp]; a[f]=1 else gotos[tmp] = {f,1} end
      end

      i = 0
      until gotos.empty? do
        i += 1

        gotos.each do |goto, froms|
          next if lock[goto.ident]
          lock[goto.ident] = true
          puts "la: goto #{goto.ident}" if d_reduce
          goto.first_term.each {|i| ret[i.ident] = i }

  goto.reduce_core.each do |pt|
    sim = pt.rule.symbol
    froms.each do |f,len|
      if pt.index == len then
                  tmp = f.goto_table[ sim ]
                  #unless tmp then p goto; p f; p pt; bug! end
                  if h = new[tmp]; h[f]=1 else new[tmp] = {f,1} end
      else
        ff = f
        #begin
        ff.head_state( pt.before(len) ).each_key do |f|
                  tmp = f.goto_table[ sim ]
                  #unless tmp then p goto; p f; p pt; bug! end
                  if h = new[tmp]; h[f]=1 else new[tmp] = {f,1} end
        end
        #rescue FindBug
        #p f; p len; p pt; raise
        #end
      end
    end
  end
  goto.void.each do |nt|
                  tmp = goto.goto_table[ nt ]
                  #unless tmp then p goto.ident; p nt; bug! end
                  unless h = new[tmp] then
                    new[tmp] = h = {}
                  end
                  froms.each {|f,len| bug! if h[f]; h[f] = len + 1 }
  end
        end          # gotos.each

        tmp = gotos
        gotos = new
        new = tmp
        new.clear
      end            # until

      st_end = Time.times.utime
      if @profile then
        printf "%-4d %4d %4d %f\n",
               @ident, lock.nitems, i, st_end - st_beg
      end

      puts "la< state #{@ident}" if @d_reduce
      ret
    end


    def lookahead__new( ptr )
      sim = ptr.rule.symbol
      ret = []
      lock = []

      head_state( ptr ).each do |f|
        st = f.goto_table[ sim ]
        unless lock[st.ident] then
          upd_la( ret, st, f, 1, lock )
        end
      end

      ret
    end

    def upd_la( ret, st, from, len_to_from, lock )
      lock[st.ident] = true
      i = nil
      st.first_term.each {|i| ret[i.ident] = i }

      tmp = f = nil

      st.reduce_core.each do |pt|
        sim = pt.rule.symbol

        if pt.index == len_to_from then
          tmp = from.goto_table[ sim ]
          upd_la( ret, tmp, from, 1, lock ) unless lock[tmp.ident]
        else
          from.head_state( pt.before(len_to_from) ).each do |f|
            tmp = f.goto_table[ sim ]
            upd_la( ret, tmp, f, 1, lock ) unless lock[tmp.ident]
          end
        end
      end
      st.void.each do |nt|
        tmp = st.goto_table[ nt ]
        upd_la( ret, tmp, from, len_to_from + 1, lock ) unless lock[tmp.ident]
      end
    end


    def head_state( ptr )
      puts "hs> ptr #{ptr}" if @d_reduce

      cur = [self]
      new = []
      backed = st = s = nil

      until ptr.head? do
        ptr = ptr.decrement
        tok = ptr.unref

        cur.each do |st|
          unless backed = st.from_table[ tok ] then
            bug! "from table void: state #{st} key #{tok}"
          end
          backed.each {|s| new[s.ident] = s }
        end

        tmp = cur
        cur = new
        new = tmp
        cur.compact!
        new.clear
      end

      puts "hs< backed [#{sh2s cur}]" if @d_reduce
      cur
    end

    def sh2s( sh )
      sh.collect{|s,v| s.ident }.join(',')
    end


    #
    # Shift/Reduce Conflict resolver
    #

    def resolve_sr

      @goto_table.each do |stok, goto|
        next unless stok.terminal?

        unless act = @action[ stok ] then
          # no conflict
          @action[ stok ] = @actions.shift( goto )
        else
          unless ReduceAction === act then
            bug! "wrong act in action table: #{act}(#{act.type})"
          end

          # conflict on stok

          rtok = act.rule.prec
          ret  = do_resolve_sr( stok, rtok )

          case ret
          when :Reduce        # action is already set

          when :Shift         # overwrite
            act.discard
            @action[ stok ] = @actions.shift( goto )

          when :Remove        # remove
            act.discard
            @action.delete stok

          when :CantResolve   # shift as default
            act.discard
            @action[ stok ] = @actions.shift( goto )
            sr_conflict stok, act.rule
          end
        end
      end
    end
    
    def do_resolve_sr( stok, rtok )
      puts "resolve_sr: s/r conflict: rtok=#{rtok}, stok=#{stok}" if @d_shift

      unless rtok and rtok.prec then
        puts "resolve_sr: no prec for #{rtok}(R)" if @d_shift
        return :CantResolve
      end
      rprec = rtok.prec

      unless stok and stok.prec then
        puts "resolve_sr: no prec for #{stok}(S)" if @d_shift
        return :CantResolve
      end
      sprec = stok.prec

      ret = if rprec == sprec then
              case rtok.assoc
              when :Left     then :Reduce
              when :Right    then :Shift
              when :Nonassoc then :Remove  # 'rtok' makes error
              else
                bug! "#{rtok}.assoc is not Left/Right/Nonassoc"
              end
            else
              if rprec > sprec
              then :Reduce
              else :Shift
              end
            end

      puts "resolve_sr: resolved as #{ret.id2name}" if @d_shift
      ret
    end
    private :do_resolve_sr



    def rr_conflict( high, low, ctok )
      if @method == :slr then
        raise SLRerror, "SLR r/r conflict in state #{@ident}"
      end
        
      c = RRconflict.new( @ident, high, low, ctok )

      unless @rrconf then
        @rrconf = {}
      end
      if a = @rrconf[ctok] then
        a.push c
      else
        @rrconf[ctok] = [c]
      end
    end

    def sr_conflict( shift, reduce )
      if @method == :slr then
        raise SLRerror, "SLR s/r conflict in state #{@ident}"
      end

      c = SRconflict.new( @ident, shift, reduce )

      unless @srconf then
        @srconf = {}
      end
      if a = @srconf[shift] then
        a.push c
      else
        @srconf[shift] = [c]
      end
    end

  end   # LALRstate



  class LALRactionTable

    def initialize( rl, st )
      @ruletable = rl
      @statetable = st

      @reduce = []
      @shift = []
      @accept = nil
      @error = nil
    end


    def reduce_n
      @reduce.size
    end

    def reduce( i )
      if Rule === i then
        i = i.ident
      else
        i.must Integer
      end

      if ret = @reduce[i] then
        ret.used
      else
        @reduce[i] = ret = ReduceAction.new( @ruletable[i] )
      end

      ret
    end

    def each_reduce( &block )
      @reduce.each &block
    end


    def shift_n
      @shift.size
    end

    def shift( i )
      if LALRstate === i then
        i = i.ident
      else
        i.must Integer
      end

      unless ret = @shift[i] then
        @shift[i] = ret = ShiftAction.new( @statetable[i] )
      end

      ret
    end

    def each_shift( &block )
      @shift.each &block
    end


    def accept
      unless @accept then
        @accept = AcceptAction.new
      end

      @accept
    end

    def error
      unless @error then
        @error = ErrorAction.new
      end

      @error
    end

  end


  class LALRaction
  end


  class ShiftAction < LALRaction

    def initialize( goto )
      goto.must LALRstate
      @goto_state = goto
    end

    attr :goto_state

    def goto_id
      @goto_state.ident
    end

    def inspect
      "<shift #{@goto_state.ident}>"
    end

  end


  class ReduceAction < LALRaction

    def initialize( rule )
      rule.must Rule
      @rule = rule
      @refn = 1
    end

    attr :rule
    attr :refn

    def ruleid
      @rule.ident
    end

    def inspect
      "<reduce #{@rule.ident}>"
    end

    def used
      @refn += 1
    end

    def discard
      @refn -= 1
      if @refn < 0 then
        bug! 'act.refn < 0'
      end
    end

  end


  class AcceptAction < LALRaction

    def inspect
      "<accept>"
    end

  end


  class ErrorAction < LALRaction

    def inspect
      "<error>"
    end
  
  end


  class ConflictData ; end

  class SRconflict < ConflictData

    def initialize( sid, shift, reduce )
      @stateid = sid
      @shift   = shift
      @reduce  = reduce
    end
    
    attr :stateid
    attr :shift
    attr :reduce
  
    def to_s
      sprintf( 'state %d: S/R conflict rule %d reduce and shift %s',
               @stateid, @reduce.ruleid, @shift.to_s )
    end

  end

  class RRconflict < ConflictData
    
    def initialize( sid, high, low, tok )
      @stateid   = sid
      @high_prec = high
      @low_prec  = low
      @token     = tok
    end

    attr :stateid
    attr :high_prec
    attr :low_prec
    attr :token
  
    def to_s
      sprintf( 'state %d: R/R conflict with rule %d and %d on %s',
               @stateid, @high_prec.ruleid, @low_prec.ruleid, @token.to_s )
    end

  end

end   # module Racc
